//////////////////////////////////////
//File: computeRoburstNormalForm.cc
//The front-end to the robust
//normal form computations
/////////////////////////////////////

//Author: Tomas Johnson
/////////////////////////////////////
#include "computeRobustNormalForm.h"


bool computeRobustNormalForm(interval lambda_s,
			     interval lambda_u,
			     dim2taylor (*F_s)(dim2taylor_vector),
			     dim2taylor (*F_u)(dim2taylor_vector),
			     int order,
			     cinterval (*F_s_cplx)(civector), 
			     cinterval (*F_u_cplx)(civector), 
			     bool isPolynomial,
			     real F_radiusOfConvergence,
			     real &r_2 , 			     	       
			     real &r_3,			    
			     real &K_0,
			     real &kappa
			     ) {
  //--------------------------------------
  //Set the parameters,
  //Read from 'config.txt'

  int l_requested, maxOrder, pre_N_G;
  real eta, mu, eps_phi, eps_G, eps;
  ifstream InFile("config.txt", ios_base::in);
  InFile>>l_requested>>maxOrder>>pre_N_G
	>>eta>>mu>>eps_phi>>eps_G>>eps;
  InFile.close();
  if(order>maxOrder)
    maxOrder=order;
  //--------------------------------------

  //--------------------------------------
  //Generate the field

  OrgField F(lambda_s, lambda_u, order, maxOrder, eta, 
	     l_requested, F_s, F_u, 
	     F_s_cplx, F_u_cplx, isPolynomial,
	     F_radiusOfConvergence);
  cout<<"F generated"<<endl;
  //--------------------------------------

  //--------------------------------------
  //Generate the change of coordinates
  //--------------------------------------

  Phi P(F, eta, mu);
  cout<<"P generated"<<endl;

  //--------------------------------------
  //Prove that the change of variables
  //converges and set output

  F.generateA(Sup(P.get_candidateRadius()));
  P.proveConvergence(F);
  cout<<"A = "<<F.get_A()<<endl; 
  P.setOutputVariables(eps_phi);
  if(Sup(P.get_K_0()*P.get_r_phi())>=1.0) {
    eps_phi=Inf(interval(1.0)/(P.get_K_0()*P.get_r_0()) );
    P.setOutputVariables(eps_phi);
    cout<<"New epsilon_phi (norm bound on the inverse change of coordinates (possibly) invalid"<<endl
	<<"eps_phi = "<<eps<<endl;
  }

  cout<<"Phi is analytic on the disk with radius = "
      <<P.get_r_0()<<endl
      <<"K_0 <= "<<P.get_K_0()
      <<" on the disk r_0 = "<<P.get_r_phi()<<endl;
  //--------------------------------------

  //--------------------------------------
  //Generate the perturbed field
  int N_G=min(pre_N_G*F.get_l(),F.get_expOrder());
  PertField G(N_G, F, P);
  cout<<"G generated"<<endl;
  //--------------------------------------

  //--------------------------------------
  //Prove that the perturbed field
  //converges and set output
  
  G.proveConvergence(F,P);
  G.setOutputVariables(F, eps_G, eps);
  if(Sup(G.get_r_3())>=Inf(P.get_r_phi()*(interval(1.0)-P.get_K_0()*P.get_r_phi())) ) {
    eps=Inf((P.get_r_phi()*((interval(1.0)-P.get_K_0()*P.get_r_phi())/G.get_r_2())*(real)0.95));
    G.setOutputVariables(F, eps_G, eps);
    cout<<"New epsilon (norm bound on the inverse change of coordinates (possibly) invalid"<<endl
	<<"eps = "<<eps<<endl;
  }
  if(Sup(G.get_r_3()*(interval(1.0)+P.get_K_0()*G.get_r_3()))>=Inf(G.get_r_2())) {
    eps=Inf((interval(-0.5)+sqrt(interval(0.25)+G.get_r_2()*P.get_K_0() ))/(G.get_r_2()*P.get_K_0())*(real)0.95);
    G.setOutputVariables(F, eps_G, eps);
    cout<<"New epsilon (change of coordinates (possibly) outside of G's domain."<<endl
	<<"eps = "<<eps<<endl;
  }
  cout<<"G analytic on the disk with radius = "
      <<G.get_r_1()<<endl
      <<"kappa <= "<<G.get_kappa()
      <<" on the disk r_2 = "<<G.get_r_2()<<endl
      <<"We recommend that you change to the normal form "<<endl
      <<"on the disk with radius r_3 = "<<G.get_r_3()<<endl
      <<"Inverse transformation is valid inside the disk with radius = "<<P.get_r_phi()*(interval(1.0)-P.get_K_0()*P.get_r_phi())<<endl<<endl;
  
  //--------------------------------------

  //--------------------------------------
  //Set the values of the return variables
  
  r_2=Inf(G.get_r_2());
  r_3=Inf(G.get_r_3());
  kappa=Sup(G.get_kappa());
  K_0=Sup(P.get_K_0());
  return true;
  //--------------------------------------
}
