/////////////////////////////////////
//File: fieldsAndCoord.cc
/////////////////////////////////////
//Contains all data structures and 
//functions that are needed for 
//the robust normal form computations
/////////////////////////////////////

//Author: Tomas Johnson
/////////////////////////////////////


#include "fieldsAndCoord.h"

/////////////////////////////////////////////////////////////////////////
//class Phi: THE COORDINATE CHANGE !!!
/////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------
//Initialize a 2-dimensional AD-vector variable 
//with 'order'-derivatives at zero

dim2taylor_vector zeroExp(int order){
  ivector x(2);
  x[1]=0.0;
  x[2]=0.0;
  dim2taylor_vector xx;
  xx=init_var(order,x);
  return xx;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute the Taylor expansion of Phi_u
//and Phi_s up to order OrgField::expOrder

void Phi::computePhi_us(const OrgField &F) {
  //Initialise as the zero series

  phi_s=init_const(2, interval(0.0));
  phi_u=init_const(2, interval(0.0));

  //Use x as a seed for the expansion
  dim2taylor_vector x;

  //y is the argument at a given stage
  dim2taylor_vector y;

  for(int i=2; i<=F.get_expOrder(); i++) {
    //Phi_i is computed using F_i(x+phi)
   
    x=zeroExp(i);
    y[1]=x[1]+phi_s;
    y[2]=x[2]+phi_u;

    //Pre-values for phi_s, and phi_u
    phi_s=F.F_s(y);
    phi_u=F.F_u(y);

    //F_i has at least order 2, 
    //so phi_i[0]=phi_i[1]=0
    phi_s.rescale(0,0,interval(0.0));
    phi_u.rescale(0,0,interval(0.0));
    phi_s.rescale(0,1,interval(0.0));
    phi_u.rescale(0,1,interval(0.0));
    phi_s.rescale(1,0,interval(0.0));
    phi_u.rescale(1,0,interval(0.0));
    for(int k=2; k<=i; k++) {

      //Rescale the pre-values of 
      //phi_s and phi_u
      for(int j=0; j<=k; j++) {
	
	//If m \in V_l:
	//Rescale with |m\Lambda-\lambda_i|
	if(j<F.get_l() || (k-j) < F.get_l() ) {
	  phi_s.rescale(k-j,j,(real)1.0/F.get_divisor(k-j,j,1));
	  phi_u.rescale(k-j,j,(real)1.0/F.get_divisor(k-j,j,2));
	}
	
	//phi_i = [phi_i]_{V_l},
	//so any computation in U_l is removed
	else {
	  phi_s.rescale(k-j,j,interval(0.0));
	  phi_u.rescale(k-j,j,interval(0.0));
	}
      }
    }

    //Increase the expansion order
    if(i<F.get_expOrder()) {
      dim2taylor temp_s(phi_s,1);
      dim2taylor temp_u(phi_u,1);
      phi_s=temp_s;
      phi_u=temp_u;
    }
  }
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute the majorant phi_hat

void Phi::computePhi_hat(const OrgField &F){
  int order=F.get_expOrder();
  ivector x(0, order);
  x[0]=0.0;
  //We use the notation \hat \alpha_1=1.0
  x[1]=1.0;
  for(int i=2; i<=order; i++) {
    x[i]=0.0;

    //On each level sum the maximum 
    //of the stable and unstable coefficients
    for(int j=0; j<=i; j++) {
      x[i]+=max(Sup(abs(phi_s[i-j][j])),Sup(abs(phi_u[i-j][j])));
    }
  }
  phi_hat=x;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Constructor:
Phi::Phi(const OrgField &F, real eta, real mu) {
  //Make sure that eta and mu are set so that
  //n_1>N_1 and n_1>2n_0
  if(eta<0.0)
    eta=0.2;
  if((real)mu>eta)
    mu=eta/(real)2.0;

  //Compute the values of n_0 and n_1 
  //That are used for the induction step
  //in the proof of analyticity of phi_hat
  n_0=(int)floor(_double((real)F.N_l*(mu+(real)1)/(real)2));
  n_1=(int)ceil(_double((real)F.N_l*(eta+(real)1)));
  cout<<"n_0 = " <<n_0<<", n_1 = "<<n_1<<endl;

  //Compute phi_s, phi_u and phi_hat
  computePhi_us(F); 
  computePhi_hat(F);
 
  //Use the least squares routine to estimate
  //the values of C and M
  estimateCoef(get_all_coef(phi_hat), n_0+1, n_1, C, M);

  //If F is not entire, modify M,
  //so that F converges on 2/M

  if((F.radiusOfConvergence)>0.0) {
    if(F.radiusOfConvergence<Sup((real)2.0/(M))) {
      M=(real)2.0/(F.radiusOfConvergence);
    }
  }


  //To avoid mistakes, initialise the radii with
  //negative values, these are reset when 
  //converges = true

  converges=false;
  r_0=-1;
  r_phi=-1;
  K_0=-1;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Prove that Phi is an analytic function, and
//compute a lower bound on its radius of convergence

void Phi::proveConvergence(OrgField &F) {
  
  interval maxC;
  if (Sup(-F.lambda_s)<Inf(F.lambda_u))
    maxC=-F.lambda_s/F.A;
  else
    maxC=F.lambda_u/F.A;
  
  //Repeat until Phi converges
  while(!converges) {
    
    //Compute the sum in the 
    //induction criteria
    interval sum(0.0);
    for(int i=1; i<=n_0; i++)
      sum+=phi_hat[i]*power(M,-i);
  
    //Compute the induction criteria
    interval check=F.A/F.Omega(n_1+1)*((real)2*sum+(real)(n_1)*C);//-2*n_0)*C);
   
    //If check is less than one
    //the induction goes through 
    //but we prefer C<maxC

    if(Sup(check)<1.0) {
      if(Inf(maxC)<=Sup(C))
	C=(real)0.99*maxC;
      else {
	converges=true;
	//The convergence radius is 1/M
	r_0=interval(Sup((real)1.0/M));
      }
    }
    else {
      //When the proof fails, decrease
      //the radius of the disk,
      //and set C so that 
      //\hat \alpha_k<CM^k, n_0+1<k<n_1

      M*=1.2;
      ivector coef=get_all_coef(phi_hat);
      C=(real)1.0001*coef[n_0+1]/power(M,n_0+1);
      for(int i=n_0+2; i<=n_1; i++) {
	if( Inf(coef[i])>Sup(C*power(M,i)))
	  C=(real)1.0001*coef[i]/power(M,i);
      }
    }
    //Recompute A on the new disk
    F.generateA(Sup(get_candidateRadius()));   
  }
 
  cout<<"C = "<<C<<"  M= "<<M<<endl;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Given r_0, we compute K_0 on a disk with
//radius r_phi = r_0*eps_phi.
//K_0 is used to estimate the norms of phi_i 
//and their inverse. That is the norm estimates
//on phi_i are only valid on the disk with radius r_phi

void Phi::setOutputVariables(real eps_phi) {
  r_phi=r_0*eps_phi;
  interval sum(0.0);
  for(int i=2; i<=n_0; i++)
    sum+=phi_hat[i]*power(r_phi,i-2);
  K_0=sum+C*sqr(M)*power(M*r_phi,n_0-1)/((real)1.0-M*r_phi);
}
//----------------------------------------------------------



/////////////////////////////////////////////////////////////////////////
//class OrgField: THE ORIGINAL FIELD !!!
/////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------
//Find a l<=l_requested, such that \Lambda\in F_l

bool OrgField::computel(int l_requested) {
  interval su_quot(lambda_s/lambda_u);
  interval us_quot(lambda_u/lambda_s);
  l=1;
  for(int i=2; i<=l_requested; i++) {

    //Check if (l-1)*lambda_i/lambda_j is 
    //an integer

    if(ceil(_double(Inf(su_quot*(real)(i-1))))
       ==floor(_double(Sup(su_quot*(real)(i-1)))) ||
       ceil(_double(Inf(us_quot*(real)(i-1))))
       ==floor(_double(Sup(us_quot*(real)(i-1)))) )
      
      return true;

    l=i;
  }
  return false;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute the N_l associated with l and \Lambda
void OrgField::computeN_l() {
  if(Inf((-lambda_s))>Sup(lambda_u))
    N_l=l+(int)ceil(_double(Sup(abs(lambda_s/lambda_u)*(real)(l-1))));
  else
    N_l=l+(int)ceil(_double(Sup(abs(lambda_u/lambda_s)*(real)(l-1))));
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute the majorant F_hat
void OrgField::computeF_hat(int order){
  dim2taylor_vector xx=zeroExp(order);
  dim2taylor Fs=F_s(xx);
  dim2taylor Fu=F_u(xx); 
  ivector x(0, Fs.get_p());
  for(int i=0; i<=Fs.get_p(); i++) {
    x[i]=interval(0.0);

    //On each level sum the maximum 
    //of the stable and unstable coefficients

    for(int j=0; j<=i; j++) {
      x[i]+=max(Sup(abs(Fs[i-j][j])),Sup(abs(Fu[i-j][j])));
    }
    F_hat=x;
  }
}
//----------------------------------------------------------


//----------------------------------------------------------
//Constructor of the Original Field
OrgField::OrgField(interval lambda_s_in, interval lambda_u_in, 
		   int order, int maxOrder,
		   real eta, int l_requested,
		   dim2taylor (*F_s_in)(dim2taylor_vector), 
		   dim2taylor (*F_u_in)(dim2taylor_vector),
		   cinterval (*F_s_cplx_in)(civector), 
		   cinterval (*F_u_cplx_in)(civector),
		   bool isPolynomial_in,
		   real radiusOfConvergence_in) {

  //Use the given field, in the form:
  //diag(\lambda_s, \lambda_u)x+(F_s(x),F_u(x))

  lambda_s=lambda_s_in;
  lambda_u=lambda_u_in;
  F_s=F_s_in;
  F_u=F_u_in;
  F_s_cplx=F_s_cplx_in;
  F_u_cplx=F_u_cplx_in;
  isPolynomial = isPolynomial_in;
  radiusOfConvergence=radiusOfConvergence_in;
  
  //Check for resonances, and find
  //l, so that \Lmabda\in F_l.
  //If no resonances are found,
  //l=l_requested.

  bool resonance;
  do{
    resonance=computel(l_requested);
    computeN_l();
    if((real)N_l*((real)1+eta)<maxOrder)
      break;
  }
  while(l_requested--);
  if(resonance) {
    cout<<"Resonance of order "<<l<<" detected"<<endl;
    if(l==1){
      cout<<"Simplification not possible !"<<endl;
      exit(1);
    }
  }
  if(l==1) {
    cerr<<"Not able to compute any normal form."<<endl
      <<"You have to increase the maximal order in 'config.txt'"<<endl
	<<"which is currently set to, maxOrder = "<<maxOrder<<endl; 
    exit(1); 
  }
  cout<<"l = "<<l<<endl<<"N_l = "<<N_l<<endl;

  //Compute the maximal ordered needed for the computations

  int sizeOf_n_1=(int)ceil(_double((real)N_l*((real)1.0+eta)));
  order=max(order, sizeOf_n_1);
  computeF_hat(order);
  expOrder=order;
  cout<<"Order of taylor approximations = "<<expOrder<<endl;

  //s and A cannot be initialised until a coordinate change
  //Phi is known. 
  s=-1;
  A=-1;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute \Omega(n), the minimal divisor at level n,
//that is, min(|m\Lambda-\lambda_i| : m\in V_l, |m|=n)

interval OrgField::Omega(int n) const{
  if (Sup(-lambda_s)<Inf(lambda_u))
    return abs((real)(n-l)*abs(lambda_s)+(real)(l-1)*lambda_u);
  else  
    return abs((real)(n-l)*abs(lambda_u)+(real)(l-1)*lambda_s);
    
}
//----------------------------------------------------------


//----------------------------------------------------------
//Compute A, such that F_hat(r) <= Ar^2.
//On the disk |x|<s

void OrgField::generateA(real s_in) {
  s=s_in;
  A=0.0;
  for(int i=2; i<=get_order(F_hat); i++){
    A+=F_hat[i]*power(s,i-2);
  }

  //If F is not a polynomial
  //Use Cauchy estimates on the 
  //higher order derivatives

  if(!isPolynomial) {
    interval rho(s*interval(-2,2));
    cinterval rhoBox(rho,rho);
    civector z(2);
    z[1]=rhoBox;
    z[2]=rhoBox;
    interval normBound(Sup(abs(F_s_cplx(z)))+Sup(abs(F_u_cplx(z))));
    A+=normBound/sqr(s)/power((real)2, expOrder)*((real)(expOrder+3));
  }
    

}
//----------------------------------------------------------


//////////////////////////////////////////////////////////////////////////////////

/////////////////////////////////////////////////////////////////////////
//class PertField: THE PERTURBED FIELD !!!
/////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------
//Compute the majorant \hat g

void PertField::computeg_hat(const ivector &F_hat, 
			     itaylor &phi_hat,
			     int start, int stop) {

  //Initialise as zero
  g_hat=const_itaylor(get_order(phi_hat), interval(0.0));
  itaylor firstTerm;

  //g_hat is given by a sum, 
  //where the first term is F_hat(x+phi_hat), 
  
  firstTerm=const_itaylor(get_order(phi_hat), interval(0.0));
  
  //the second term involves the derivative of phi_hat
  ivector phi_prim_v(0,get_order(phi_hat));
  phi_prim_v[0]=interval(0.0);

  //Compute the first term, and compute the 
  //derivative of phi_hat

  for(int i=1;i<=get_order(phi_hat); i++) {
    firstTerm=firstTerm+F_hat[i]*powerAtZero(phi_hat, i);
    if(i<get_order(phi_hat))
      phi_prim_v[i]=(real)(i+1)*phi_hat[i+1];
  }

  //Initialise the derivative of phi_hat
  //as an AD variable.

  itaylor phi_prim;
  phi_prim=phi_prim_v;

  //Compute g_hat

  for(int i=start; i<=stop; i++) {
    g_hat=firstTerm+(real)2*phi_prim*g_hat;
    ivector temp=get_all_coef(g_hat);
   
    //Since G=[G]_{U_l}, 
    //the first (2l-1) term can be removed

    for(int j=0; j<start; j++)
      temp[j]=0.0;

    g_hat=temp;
  }
}
//----------------------------------------------------------


//----------------------------------------------------------
//Constructor of the perturbed field

PertField::PertField(int N_G_in, const OrgField &F, Phi &P) {
  //Set N_G
  N_G=N_G_in;

  //Extract the coefficients of F.hat,
  //up to order OrdField::expOrder

  ivector F_hat_terms=get_all_coef(F.F_hat);
  ivector F_hat_terms_argument=F_hat_terms(0,F.get_expOrder());

  //Compute the first (pre_N_G-2)l non-zero terms of G_hat

  computeg_hat(F_hat_terms_argument,P.phi_hat, 2*F.l, N_G);

  //Use the first few g_hat to esetimate D and K

  estimateCoef(get_all_coef(g_hat), 2*F.l, N_G, D, K);

  //G has a smaller radius of convergence than 
  //Phi, we can therefore assume that K>M

  if (Sup(K)<Inf(P.M))
    K=(real)1.02*P.M;
  
  //To avoid mistakes, initialise the radii with
  //negative values, these are reset when 
  //converges = true
  
  converges=false;
  r_1=-1;
  r_2=-1;
  r_3=-1;
  kappa=-1;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Prove that G is an analytic function, and
//compute a lower bound on its radius of convergence

void PertField::proveConvergence(const OrgField &F, Phi &P) {

  //The condition for the induction step
  //involves two terms, \Sigma_1, and \Sigma_2.
  //Each term contains a sum, 
  //these are computed first.

  interval sum1(interval(1.0)/P.M), sum2(0.0);
  for(int i=2; i<=P.n_0; i++) {
    sum1+=P.phi_hat[i]*power(P.M,-i);
    sum2+=(real)i*P.phi_hat[i]*power(K,1-i);
  } 
  
  while(!converges) { 

    //check1 is A*\Sigma_1

    interval check1=F.A*( (real)2*sum1+(real)(N_G+1-2*P.n_0)*P.C )*P.C/D*power(P.M/K,N_G+2);

    //check2 is 2*\Sigma_2
    
    interval check2=(real)2*( sum2+P.C*P.M*power(P.M/K,P.n_0)*
			      ( (real)(P.n_0+1)-(real)(P.n_0)*(P.M/K) )
			      /sqr(interval(1.0)-P.M/K) );

    //Check if the condition for the induction step holds
    if(Sup(check1+check2)<1.0) {
      converges=true;
      r_1=interval(Sup((real)1.0/K));
    }

    //If not convergent, increase either D or M
    //to increase M, decreases both \Sigma_i
    //whereas an increment of D only decreases
    //\Sigma_1.
    //However we want M as large as possible,
    //therefore we increase D if \Sigma_1
    //is the dominant term

    else if(Inf(check1)>Sup(check2)) {
      D*=2.0;
    }
    else {
      sum2=0.0;
      K*=1.05;
      
      //sum2 has to be recomputed

      for(int i=2; i<=P.n_0; i++) {
	sum2+=(real)i*P.phi_hat[i]*power(K,1-i);
      }   
    }
  }
  cout<<"D = "<<D<< " K = "<<K<<endl;
}
//----------------------------------------------------------


//----------------------------------------------------------
//Given r_1, we compute \kappa on a disk with
//radius r_2 = r_1*eps_G.
//kappa is used to estimate the norms of G.
//That is the norm estimates
//on G are only valid on the disk with radius r_2.
//When the variables are changed, some room must be left
//for the correction due to the norm of phi.
//Therefore the flow should only be interrupted
//on a smaller disk, with radius r_3. 
//So that all calcutlation are done within 
//the disk of radius min(r_phi,r_2).
//This is ensured by taking r_3=eps*r_2,
//with eps small enough.

void PertField::setOutputVariables(const OrgField &F, real eps_G, real eps) {
  r_2=r_1*eps_G;
  r_3=r_2*eps;

  //Kappa yields the estimates on the flow and the flow-times.
  kappa=D*power(K,2*F.l)/((real)1.0-K*r_2)*power(r_2,2*F.l-1);
}
//----------------------------------------------------------


//////////////////////////////////////////////////////////////////////////////////
//Auxiliary Functions
//////////////////////////////////////////////////////////////////////////////////

//----------------------------------------------------------
//A least squares estimator of the bounds on the coefficients 
//of an analytic function.
//Also ensures that all coefficients in the range (start,stop)
//are bounded by factor*exponent^k

void estimateCoef(const ivector &coef, int start, int stop, interval &factor, interval &exponent) {
  rmatrix A(stop-start+1,2);
  rvector rhs(stop-start+1);
 
  //Construct a the coefficient matrix,
  //and the right hand side
  for(int i=1; i<=(stop-start+1); i++) {
    A[i][1]=real(1.0);
    A[i][2]=real(i+start-1);
    rhs[i]=ln(mid(coef[i+start-1]));
  }

  //Reduce to a quadratic system
  rmatrix lhs=transp(A)*A;
  rhs=transp(A)*rhs;
  rmatrix preCond(4);
  int err;
  MatInv(lhs,preCond,err);

  //Solve the quadratic system
  rvector sol=preCond*rhs;
  factor=exp(sol[1]);
  exponent=exp(sol[2]);

  //Ensure that the all coefficients used
  //to estimate the bound are actually
  //bounded by it.
  //We assume that the linear term is ok
  //and increase (if necessary)
  //the constant one.

  for(int i=start; i<=stop; i++) {
    if( Inf(coef[i])>Sup(factor*power(exponent,i)))
      factor=(real)1.0001*coef[i]/power(exponent,i);
  }
}
//----------------------------------------------------------


//----------------------------------------------------------
//Computes powers of a 2-dimensional AD vector
//without errors at zero
dim2taylor POWER(const dim2taylor &z, const int &n) {
  if (n==1)
    return z;
  if (n==2)
    return sqr(z);
  int k=n/2;
  if (n==2*k)
    return sqr(POWER(z,k));
  else
    return z*sqr(POWER(z,k));
}
//----------------------------------------------------------


//////////////////////////////////////////////////////////////////////////////////////////
