/////////////////////////////////////
//File: fieldsAndCoord.h
/////////////////////////////////////
//Contains all data structures and 
//functions that are needed for 
//the robust normal form computations
/////////////////////////////////////

//Author: Tomas Johnson
/////////////////////////////////////


#ifndef __fieldsAndCoord_h__
#define __fieldsAndCoord_h__

#include <fstream>
#include <cstdlib>
#include <iostream>
#include <string>
#include <stdexcept>

#include "imath.hpp"
#include "rmath.hpp"
#include "interval.hpp"
#include "ivector.hpp"
#include "imatrix.hpp"
#include "matinv.hpp"
#include "itaylor.hpp"
#include "dim2taylor.hpp"
#include "cinterval.hpp"
#include "cimath.hpp"
#include "civector.hpp"


using namespace std;
using namespace cxsc;
using namespace taylor;

class OrgField;
class PertField;

//----------------------------------------------------------
//Two auxiliary functions:
//Estimate bounds by a least-squares approximation:
void estimateCoef(const ivector &coef, int start, int stop, 
		  interval &factor, interval &exponent);
//Compute the power function for 2-dimensional 
//taylor series at zero:
dim2taylor POWER(const dim2taylor &z, const int &n);
//----------------------------------------------------------

//---------------------------------------------------
//The change of variables class
class Phi{
 private:
  int n_0,n_1;
  dim2taylor phi_s, phi_u;
  itaylor phi_hat;
  interval C,M;
  bool converges;
  interval r_0, r_phi, K_0;

  //Compute the first OrgField::expOrder 
  //coefficients of phi_s and phi_u.
  //Use them to compute the first
  //OrgField::expOrder coefficients
  //for phi_hat.

  void computePhi_us(const OrgField &F);
  void computePhi_hat(const OrgField &F);

 public:
  Phi() {}
  Phi(const OrgField &F, real eta=0.2, real mu=-0.65);

  //Finds C and M such that Phi converges
  void proveConvergence(OrgField &F);

  //Use eps_phi to set the output
  void setOutputVariables(real eps_phi=0.5);

  inline interval get_candidateRadius() const {return interval(1.0)/M;}
  inline interval get_r_0() const {return r_0;}
  inline interval get_r_phi() const {return r_phi;}
  inline interval get_K_0() const {return K_0;}
  inline bool isAnalytic() const {return converges;}
  
  friend class PertField;
  friend void estimateCoef(const ivector &coef, int start, int stop, interval &factor, interval &exponent);
};
//---------------------------------------------------

//---------------------------------------------------
//The original field class
class OrgField{
 private:
  interval lambda_s, lambda_u;
  int l, N_l;
  dim2taylor (*F_s)(dim2taylor_vector);
  dim2taylor (*F_u)(dim2taylor_vector);
  cinterval (*F_s_cplx)(civector);
  cinterval (*F_u_cplx)(civector);
  bool isPolynomial;
  real radiusOfConvergence;
  itaylor F_hat;
  interval s, A;
  int expOrder;
  
  //Find a l <= l_requested such that
  //diag(lambda_s,lambda_u) \in F_l
  bool computel(int l_requested);

  //Compute the corresponding 
  //value of N_l
  void computeN_l();

  //Compute a majorising series 
  //of F_s and F_u
  void computeF_hat(int order);

 public:
  OrgField() {}
  OrgField(interval lambda_s_in, interval lambda_u_in, int order, int maxOrder,
	   real eta, int l_requested,
	   dim2taylor (*F_s_in)(dim2taylor_vector), dim2taylor (*F_u_in)(dim2taylor_vector),
	   cinterval (*F_s_cplx_in)(civector), cinterval (*F_u_cplx_in)(civector),
	   bool isPolynomial_in, real radiusOfConvergence_in);

  //Compute the lower bound on |m*\Lambda-\lambda_i|
  //for m \in V_l, with |m|=n.
  interval Omega(int n) const;

  void generateA(real s_in);

  inline interval get_A() const {return A;}
  inline int get_l() const {return l;}
  inline int get_expOrder() const {return expOrder;}

  //Compute the divisor |m\Lambda-\lambda_i|
  //used in the generation of phi_i
  inline interval get_divisor(int ms, int mu, int i) const 
    {if(i==1) return ((real)(ms-1)*lambda_s+(real)mu*lambda_u);
      else  return ((real)ms*lambda_s+(real)(mu-1)*lambda_u); }

  friend class Phi;
  friend class PertField;
};
//---------------------------------------------------

//---------------------------------------------------
//The perturbed field class
class PertField{
 private:
  int N_G;
  itaylor g_hat;
  interval K,D;
  bool converges;
  interval r_1, r_2, r_3, kappa;

  //Compute the first (stop+1-start)
  //non-zero components of g_hat
  void computeg_hat(const ivector &F_hat, itaylor &phi_hat, 
		    int start, int stop);
 public:
  PertField() {}
  PertField(int N_G_in, const OrgField &F, Phi &P);

  //Find D and K such that G converges
  void proveConvergence(const OrgField &F, Phi &P);

  //Use eps_G and eps to set the output
  void setOutputVariables(const OrgField &F, real eps_G=0.5, real eps=0.5);

  inline bool isAnalytic() const {return converges;}
  inline interval get_r_1() {return r_1;}
  inline interval get_r_2() {return r_2;}
  inline interval get_r_3() {return r_3;}
  inline interval get_kappa() {return kappa;}

  friend void estimateCoef(const ivector &coef, int start, int stop, 
			   interval &factor, interval &exponent);
};
//---------------------------------------------------

#endif
